from ._version import __version__
from .client import Client
from .errors import ApiError, NotFoundError
from .models import (
    Model,
    ModelGroup,
    ModelInputSpecification,
    ModelRun,
    ModelRunMode,
    ModelRunOutputData,
    ModelRunOutputResult,
    ModelRunPage,
    ModelRunStatus,
    ModelRunWithOutput,
    ModelSummary,
    Organisation,
    OutputVariable,
    Sensitivity,
    User,
)

__all__ = [
    "__version__",
    "Client",
    "ApiError",
    "NotFoundError",
    "Model",
    "ModelGroup",
    "ModelInputSpecification",
    "ModelRun",
    "ModelRunMode",
    "ModelRunOutputData",
    "ModelRunOutputResult",
    "ModelRunPage",
    "ModelRunStatus",
    "ModelRunWithOutput",
    "ModelSummary",
    "Organisation",
    "OutputVariable",
    "Sensitivity",
    "User",
]



from __future__ import annotations

from enum import StrEnum

from pydantic import BaseModel, ConfigDict, Field
from pydantic.alias_generators import to_camel


class Pd4castrBaseModel(BaseModel):
    model_config = ConfigDict(
        extra="ignore",
        alias_generator=to_camel,
        populate_by_name=True,
    )


class FileFormat(StrEnum):
    json = "json"
    csv = "csv"
    parquet = "parquet"


class ModelRunMode(StrEnum):
    automatic = "AUTOMATIC"
    on_demand = "ON_DEMAND"


class TimeHorizon(Pd4castrBaseModel):
    name: str
    short_name: str
    key: str


class ModelMetadata(Pd4castrBaseModel):
    description: str | None = None
    release_date: str | None = None
    resolution: str | None = None


class UISpecification(Pd4castrBaseModel):
    series_key: bool | None = None
    colours: list[str] | None = None


class ColumnSpecification(Pd4castrBaseModel):
    name: str
    key: str | None = None
    type: str
    ui: UISpecification | None = None


class OutputVariable(Pd4castrBaseModel):
    name: str
    key: str
    type: str
    series_key: bool | None = None
    colours: list[str] | None = None


class SensitivitySpecification(Pd4castrBaseModel):
    id: str
    name: str
    key: str


class ForecastVariableSpecification(Pd4castrBaseModel):
    name: str
    short_name: str
    key: str


class ModelInputSpecification(Pd4castrBaseModel):
    id: str
    key: str


class Model(Pd4castrBaseModel):
    id: str
    model_group_id: str
    name: str
    display_name: str
    notes: str
    revision: int
    docker_image: str
    created_at: str
    run_mode: ModelRunMode
    horizon: TimeHorizon
    model_metadata: ModelMetadata
    output_variables: list[OutputVariable] = []
    # DEPRECATED: use output_variables
    output_specification: list[ColumnSpecification] = []
    inputs: list[ModelInputSpecification] | None = None
    tags: list[str]
    sensitivities: list[SensitivitySpecification]
    output_file_format: FileFormat
    display_timezone: str
    forecast_variable: ForecastVariableSpecification


class ModelGroup(Pd4castrBaseModel):
    id: str
    name: str
    latest_revision: int
    public: bool
    created_at: str


class ModelRunStatus(StrEnum):
    pending = "pending"
    input_files_prepared = "input_files_prepared"
    running = "running"
    completed = "completed"
    failed = "failed"


class Sensitivity(Pd4castrBaseModel):
    id: str
    name: str
    key: str


class ModelRun(Pd4castrBaseModel):
    id: str
    status: ModelRunStatus
    created_at: str
    run_datetime: str
    started_at: str | None = None
    completed_at: str | None = None
    errored_at: str | None = None
    sensitivity: Sensitivity | None = None
    error: str | None = None
    container_id: str | None = None


class ModelRunPage(Pd4castrBaseModel):
    data: list[ModelRun]
    next_cursor: str | None = None
    has_more: bool


class ModelSummary(Pd4castrBaseModel):
    id: str
    display_name: str
    display_timezone: str


class ModelRunOutputData(Pd4castrBaseModel):
    model_config = ConfigDict(extra="allow")

    forecast_datetime: str = Field(alias="forecast_datetime")


class ModelRunWithOutput(Pd4castrBaseModel):
    id: str
    run_datetime: str
    sensitivity: Sensitivity | None = None
    model: ModelSummary
    colours: dict[str, str | None]
    data: list[ModelRunOutputData]


class ModelRunOutputResult(Pd4castrBaseModel):
    run: ModelRunWithOutput
    comparison_runs: dict[str, ModelRunWithOutput]


class Organisation(Pd4castrBaseModel):
    id: str
    display_name: str
    slug: str
    domains: list[str]
    permissions: list[str]


class User(Pd4castrBaseModel):
    id: str
    email: str
    organisation: Organisation | None = None



from __future__ import annotations

from typing import Any


class Pd4castrApiError(Exception):
    pass


class AuthenticationError(Pd4castrApiError):
    pass


class NotFoundError(Pd4castrApiError):
    pass


class ApiError(Pd4castrApiError):
    status_code: int
    message: str
    body: Any | None

    def __init__(
        self,
        *,
        status_code: int,
        message: str,
        body: Any | None = None,
    ) -> None:
        super().__init__(message)
        self.status_code = status_code
        self.message = message
        self.body = body

    def __str__(self) -> str:  # pragma: no cover
        return f"{self.status_code}: {self.message}"



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Protocol, runtime_checkable

import httpx


@runtime_checkable
class TransportResponse(Protocol):
    @property
    def status_code(self) -> int: ...
    @property
    def text(self) -> str: ...
    def json(self) -> Any: ...


QueryParams = (
    dict[str, str | int | float | bool | None]
    | list[tuple[str, str | int | float | bool | None]]
    | None
)


class Transport(Protocol):
    def request(
        self,
        method: str,
        url: str,
        *,
        headers: dict[str, str] | None = None,
        params: QueryParams = None,
        json: Any | None = None,
        data: Any | None = None,
    ) -> TransportResponse: ...

    def close(self) -> None: ...


@dataclass
class HttpxSyncTransport:
    client: httpx.Client

    def request(
        self,
        method: str,
        url: str,
        *,
        headers: dict[str, str] | None = None,
        params: QueryParams = None,
        json: Any | None = None,
        data: Any | None = None,
    ) -> TransportResponse:
        return self.client.request(
            method,
            url,
            headers=headers,
            params=params,
            json=json,
            data=data,
        )

    def close(self) -> None:
        self.client.close()



from __future__ import annotations

from typing import Any

import httpx

from ._version import __version__
from .auth import ClientCredentialsAuth
from .config import (
    DEFAULT_API_URL,
    DEFAULT_AUTH0_AUDIENCE,
    DEFAULT_AUTH0_DOMAIN,
    load_config,
)
from .errors import ApiError, NotFoundError
from .models import (
    Model,
    ModelGroup,
    ModelRun,
    ModelRunOutputResult,
    ModelRunPage,
    Sensitivity,
    User,
)
from .transport import HttpxSyncTransport, QueryParams, Transport


def _normalize_base_url(base_url: str) -> str:
    return base_url.rstrip("/")


class Client:
    def __init__(
        self,
        *,
        base_url: str | None = None,
        auth0_domain: str | None = None,
        auth0_audience: str | None = None,
        client_id: str,
        client_secret: str,
        timeout: float = 30.0,
        transport: Transport | None = None,
    ) -> None:
        config = load_config()

        resolved_base_url = base_url or config.base_url or DEFAULT_API_URL
        resolved_domain = auth0_domain or config.auth0_domain or DEFAULT_AUTH0_DOMAIN
        resolved_audience = (
            auth0_audience or config.auth0_audience or DEFAULT_AUTH0_AUDIENCE
        )

        self._base_url = _normalize_base_url(resolved_base_url)
        self._auth = ClientCredentialsAuth(
            auth0_domain=resolved_domain,
            audience=resolved_audience,
            client_id=client_id,
            client_secret=client_secret,
        )

        self._transport = transport or HttpxSyncTransport(
            client=httpx.Client(timeout=timeout)
        )

    def close(self) -> None:
        self._transport.close()

    def __enter__(self) -> Client:
        return self

    def __exit__(self, exc_type: object, exc: object, tb: object) -> None:
        self.close()

    def get_models(self, *, time_horizon: str | None = None) -> list[Model]:
        params: dict[str, Any] = {}
        if time_horizon is not None:
            params["timeHorizon"] = time_horizon

        data = self._request_json("GET", "/model", params=params)
        return [Model.model_validate(item) for item in data]

    def get_model(self, model_id: str) -> Model:
        data = self._request_json("GET", f"/model/{model_id}")
        return Model.model_validate(data)

    def get_model_sensitivities(self, model_id: str) -> list[Sensitivity]:
        data = self._request_json("GET", f"/model/{model_id}/sensitivities")
        return [Sensitivity.model_validate(item) for item in data]

    def get_current_user(self) -> User:
        data = self._request_json("GET", "/user/current")
        return User.model_validate(data)

    def get_model_groups(self) -> list[ModelGroup]:
        data = self._request_json("GET", "/model-group")
        return [ModelGroup.model_validate(item) for item in data]

    def get_model_group(self, model_group_id: str) -> ModelGroup:
        data = self._request_json("GET", f"/model-group/{model_group_id}")
        return ModelGroup.model_validate(data)

    def get_model_group_models(self, model_group_id: str) -> list[Model]:
        data = self._request_json("GET", f"/model-group/{model_group_id}/models")
        return [Model.model_validate(item) for item in data]

    def get_model_runs(
        self,
        model_id: str,
        *,
        limit: int | None = None,
        cursor: str | None = None,
        after: str | None = None,
        before: str | None = None,
        status: str | list[str] | None = None,
        sensitivity: str | None = None,
    ) -> ModelRunPage:
        params: dict[str, Any] = {}
        if limit is not None:
            params["limit"] = limit
        if cursor is not None:
            params["cursor"] = cursor
        if after is not None:
            params["after"] = after
        if before is not None:
            params["before"] = before
        if status is not None:
            params["status"] = status if isinstance(status, str) else ",".join(status)
        if sensitivity is not None:
            params["sensitivity"] = sensitivity
        data = self._request_json("GET", f"/model/{model_id}/run", params=params)
        return ModelRunPage.model_validate(data)

    def get_model_run(self, model_id: str, run_id: str) -> ModelRun:
        data = self._request_json("GET", f"/model/{model_id}/run/{run_id}")
        return ModelRun.model_validate(data)

    def get_model_run_output(
        self,
        model_id: str,
        model_run_id: str,
        *,
        comparison_model: str | list[str] | None = None,
    ) -> ModelRunOutputResult:
        params: QueryParams = None
        if comparison_model is not None:
            models = (
                [comparison_model]
                if isinstance(comparison_model, str)
                else comparison_model
            )
            params = [("comparisonModel", m) for m in models]
        data = self._request_json(
            "GET",
            f"/model/{model_id}/run/{model_run_id}/output",
            params=params,
        )
        return ModelRunOutputResult.model_validate(data)

    def get_latest_model(self, model_group_id: str) -> Model:
        """Get the latest revision model in a model group."""
        models = self.get_model_group_models(model_group_id)
        if not models:
            raise NotFoundError("no models found in model group")
        return max(models, key=lambda m: m.revision)

    def get_latest_model_run(
        self,
        model_id: str,
        *,
        sensitivity: str | None = None,
    ) -> ModelRun:
        """Get the most recent completed run for a model.

        By default returns the latest base run (excluding sensitivities).
        Pass a sensitivity ID to get the latest run for that sensitivity.
        """
        resolved_sensitivity = sensitivity if sensitivity is not None else "base"
        page = self.get_model_runs(
            model_id, limit=1, status="completed", sensitivity=resolved_sensitivity
        )
        if not page.data:
            raise NotFoundError("no completed runs found for model")
        return page.data[0]

    def get_latest_model_run_output(
        self,
        model_id: str,
        *,
        sensitivity: str | None = None,
    ) -> ModelRunOutputResult:
        """Get the output from the most recent completed run for a model.

        By default returns the output from the latest base run.
        Pass a sensitivity ID to get the output for that sensitivity's latest run.
        """
        latest_run = self.get_latest_model_run(model_id, sensitivity=sensitivity)
        return self.get_model_run_output(model_id, latest_run.id)

    def _request_json(
        self,
        method: str,
        path: str,
        *,
        params: QueryParams = None,
    ) -> Any:
        url = f"{self._base_url}{path}"
        token = self._auth.get_access_token(self._transport)

        response = self._transport.request(
            method,
            url,
            headers={
                "authorization": f"Bearer {token}",
                "accept": "application/json",
                "x-sdk-version": f"python-sdk/{__version__}",
            },
            params=params,
        )

        if response.status_code == 404:
            raise NotFoundError("resource not found")

        if response.status_code >= 400:
            body: Any | None
            try:
                body = response.json()
            except Exception:
                body = response.text

            raise ApiError(
                status_code=response.status_code,
                message="request failed",
                body=body,
            )

        try:
            return response.json()
        except ValueError as e:
            raise ApiError(
                status_code=response.status_code,
                message=f"invalid JSON response: {e}",
                body=response.text,
            ) from e
